import json
from typing import Callable, Iterator

from werkzeug.exceptions import HTTPException
from werkzeug.routing import Map, Rule
from werkzeug.wrappers import Response as WerkzeugResponse

from .request import Request
from .response import Response
from .utils import endpoint as gen_endpoint


class Application:
    """Wrapper over WSGI application. This class provides simple route
    registration mechanism that allows to match code execution to request
    paths. It's the central point of an app but does not provide much more
    than that.

    Possible customizations include custom JSON encoder and decoder classes
    and possibility to specify custom request and response classes.

    :cvar json_encoder: JSON encoder class, defaults to
                        :class:`~json.JSONEncoder`
    :cvar json_decoder: JSON decoder class, defaults to
                        :class:`~json.JSONDecoder`
    :cvar request_class: request encapsulation class, defaults to
                         :class:`~kristall.request.Request`
    :cvar response_class: response encapsulation class, defaults to
                          :class:`~kristall.response.Response`
    """

    METHODS = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'HEAD', 'OPTIONS']

    json_encoder = json.JSONEncoder
    json_decoder = json.JSONDecoder
    request_class = Request
    response_class = Response

    def __init__(self):
        self.url_map = Map()
        self._resource_cache = {}
        self._before_request = []
        self._after_request = []

    def add_resource(self, path: str, resource: object):
        """Register resource under specified path. The resource is an
        instance that is expected to provide methods that correspond to HTTP
        words. If the instance provides :attr:`endpoint` attribute the it will
        be used as endpoint name, otherwise it will be registered under the
        name that is it's class dotted path.

        :param path: HTTP path which resource serves
        :type path: str
        :param resource: resource instance
        :type resource: object
        """
        resource_methods = []
        for method in self.METHODS:
            handler_name = method.lower()
            handler = getattr(resource, handler_name, None)
            if handler:
                resource_methods.append(method)
        endpoint = getattr(resource, 'endpoint', None)
        if not endpoint:
            endpoint = gen_endpoint(resource)
        self.url_map.add(Rule(path, endpoint=endpoint, methods=resource_methods))
        self._resource_cache[endpoint] = resource

    def add_before_request(self, func: Callable):
        """Add function to list of callables called before handling request.
        These functions should accept request object as parameter. These
        functions should either return None or instance of Response that will
        be returned as result of dispatch.

        :param func: callable to be called before request
        :type func: Callable
        """
        self._before_request.append(func)

    def add_after_request(self, func: Callable):
        """Add function to list of callables called after handling request.
        These functions should accept request and handler result, and return
        either None or thing that will be returned from dispatch.

        :param func: callable to be called after request
        :type func: Callable
        """
        self._after_request.append(func)

    def dispatch(self, request: Request) -> Response:
        """Dispatch and service HTTP request.

        :param request: request wrapper
        :type request: Request
        :raises MethodNotAllowed: if resource that is registered for specified
                                  path does not support HTTP method
        :return: response that is either result of request handler or an error
        :rtype: Response
        """
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            resource = self._resource_cache[endpoint]
            handler = getattr(resource, request.method.lower())
            for func in self._before_request:
                result = func(request)
                if result is not None:
                    return result
            result = handler(request, **values)
            for func in self._after_request:
                ar_result = func(request, result)
                if ar_result is not None:
                    result = ar_result
                    break
        except HTTPException as e:
            return e
        if isinstance(result, WerkzeugResponse):
            return result
        if isinstance(result, str):
            return self.response_class(result)
        return self.response_class(json.dumps(result, cls=self.json_encoder))

    def wsgi_app(self, environ: dict, start_response: Callable) -> Response:
        """Wrapper over WSGI application that exposes WSGI application for
        further wrapping with WSGI middleware.

        :param environ: WSGI environ
        :type environ: dict
        :param start_response: callable that is used to begin writing response
        :type start_response: Callable
        :return: response object
        :rtype: Response
        """
        request = self.request_class(environ, self.json_decoder)
        response = self.dispatch(request)
        return response(environ, start_response)

    def __call__(self, environ: dict, start_response: Callable) -> Iterator:
        return self.wsgi_app(environ, start_response)



import inspect
from typing import Type, Union


def endpoint(item: Union[object, Type]) -> str:
    """Endpoint generation function. Returns fully qualified class name in
    dotted notation.

    :param item: item to generate endpoint for, may be either instance or
                 class
    :type item: Union[object, Type]
    :return: fully qualified class name, suitable for use as endpoint name
    :rtype: str
    """
    if inspect.isclass(item):
        mname = item.__module__
        cqname = item.__qualname__
    else:
        mname = item.__class__.__module__
        cqname = item.__class__.__qualname__
    return f'{mname}.{cqname}'



from werkzeug.wrappers import Response as WResponse


class Response(WResponse):
    """Thin wrapper over :class:`~werkzeug.wrappers.Response`. The only
    difference is predefined response content type which is set to
    ``application/json``.
    """

    default_mimetype = 'application/json'



import json
from typing import Optional, Type, Union

from werkzeug.exceptions import RequestEntityTooLarge
from werkzeug.wrappers import Request as WRequest


class Request(WRequest):
    """Wrapper over :class:`~werkzeug.wrappers.Request` that has built in
    support for JSON content. Maximum content length is set to 4 megabytes.
    """

    MAX_CONTENT_LENGTH = 4 * 1024 * 1024  # four megabytes

    def __init__(
                self, environ: dict, populate_request: bool = True,
                shallow: bool = False, json_decoder: Optional[Type] = None,
            ):
        """Object initializer that allows setting JSON decoder class to be used
        when decoding JSON content. For complete description of other arguments
        see Werkzeug documentation for :class:`~werkzeug.wrappers.BaseRequest`.
        """
        super().__init__(environ, populate_request, shallow)
        self.decoder = json_decoder
        if self.decoder is None:
            self.decoder = json.JSONDecoder

    def get_data(
                self, cache: bool = True, as_text: bool = True,
                parse_form_data: bool = False,
            ) -> Union[str, bytes]:
        """Overwritten method that retrieves request data. Difference is that
        by default it fetches data as text. For complete description of call
        arguments see Werkzeug documentation for
        :meth:`~werkzeug.wrappers.BaseRequest.get_data`. This method raises
        :exc:`~werkzeug.exceptions.RequestEntityTooLarge` if content length
        exceeds allowed size (default is 4 megabytes).
        """
        if self.content_length > self.MAX_CONTENT_LENGTH:
            raise RequestEntityTooLarge(
                f'Request size exceeds allowed {self.MAX_CONTENT_LENGTH} bytes'
            )
        return super().get_data(cache, as_text, parse_form_data)

    def get_json(self, decoder: Optional[Type] = None) -> dict:
        """A method to retrieve JSON from request data, optionally using
        specified JSON decoder class. If not provided default decoder class
        is used.

        :param decoder: JSON decoder class, defaults to None
        :type decoder: Optional[Type], optional
        :return: request data as parsed JSON
        :rtype: dict
        """
        if decoder is None:
            decoder = self.decoder
        data = self.get_data()
        return json.loads(data, cls=decoder)



import ast
import codecs
import re
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))


def read(*parts):
    with codecs.open(path.join(here, *parts), 'r') as fp:
        return fp.read()


_version_re = re.compile(r"__version\s+=\s+(.*)")


def find_version(*where):
    return str(ast.literal_eval(_version_re.search(read(*where)).group(1)))


test_reqs = [
    'pytest',
    'pytest-cov',
    'pytest-mock',
]

dev_reqs = test_reqs + [
    'pip',
    'setuptools',
    'wheel',
    'ipython',
    'ipdb',
    'watchdog',
    'flake8',
    'flake8-builtins',
    'flake8-bugbear',
    'flake8-comprehensions',
    'flake8-pytest-style',
    'pep8-naming',
    'dlint',
    'Sphinx',
    'python-dotenv',
    'rstcheck',
]


setup(
    name='kristall',
    description='Lightweight web framework for building APIs and backends',
    version=find_version('src', 'kristall', '_version.py'),
    author='Jarek Zgoda',
    author_email='jarek.zgoda@gmail.com',
    url='https://github.com/zgoda/kristall',
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    python_requires='~=3.7',
    zip_safe=False,
    install_requires=[
        'Werkzeug'
    ],
    tests_require=test_reqs,
    extras_require={
        'dev': dev_reqs,
        'test': test_reqs,
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Typing :: Typed',
    ],
)
